"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from .utils import basic_auth_header
from .secrets import (
    get_secrets_file,
    get_credential_cache,
    read_client_secrets,
    update_client_secrets,
)



"""

Copyright (C) 2018-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.auth import read_client_secrets
from sregistry.logger import RobotNamer, bot
from sregistry.main import ApiConnection
from sregistry.utils import confirm_action
import boto3
from botocore.client import ClientError
from botocore import UNSIGNED
import json
import os

from .query import search, search_all, container_search
from .pull import pull
from .push import push
from .delete import delete


class Client(ApiConnection):
    def __init__(self, secrets=None, base=None, **kwargs):
        self.base = base
        self._update_secrets()
        self._update_headers()
        super(Client, self).__init__(**kwargs)

    def _speak(self):
        """add the bucket"""
        bot.info("[bucket:s3://%s]" % self.bucket)

    def __str__(self):
        return type(self)

    def _read_response(self, response, field="detail"):
        """attempt to read the detail provided by the response. If none,
        default to using the reason"""

        try:
            message = json.loads(response._content.decode("utf-8"))[field]
        except:
            message = response.reason
        return message

    def get_bucket_name(self):
        """get or return the s3 bucket name. If not yet defined via an environment
        variable or setting, we create a name with the pattern.
                 sregistry-<robotnamer>-<1234>

        You can use the following environment variables to determine
        interaction with the bucket:

        SREGISTRY_S3_BUCKET: the bucket name (all lowercase, no underscore)

        """
        # Get bucket name
        bucket_name = "sregistry-%s" % RobotNamer().generate()
        self.bucket_name = self._get_and_update_setting(
            "SREGISTRY_S3_BUCKET", bucket_name
        )

    def get_bucket(self):
        """given a bucket name and a client that is initialized, get or
        create the bucket.
        """
        for attr in ["bucket_name", "s3"]:
            if not hasattr(self, attr):
                bot.exit("client is missing attribute %s" % (attr))

        self.bucket = self.s3.Bucket(self.bucket_name)
        # check if the bucket exists and can be accessed
        # based on https://stackoverflow.com/a/47565719/1253230
        try:
            self.s3.meta.client.head_bucket(Bucket=self.bucket.name)
        except ClientError as e:
            self.bucket = None
            error_code = int(e.response["Error"]["Code"])
            if error_code == 403:
                if not self._id or not self._key:
                    bot.exit(
                        "The {} bucket is not public and needs AWS credentials to be accessed".format(
                            self.bucket_name
                        )
                    )
                bot.exit(
                    "The {} bucket cannot be accessed with the provided AWS credentials".format(
                        self.bucket_name
                    )
                )
            elif error_code == 404:
                # If the bucket doesn't exist, ask the user if he/she wants to try creating it
                if confirm_action(
                    "Are you sure you want to create the {} bucket?".format(
                        self.bucket_name
                    )
                ):
                    if not self._id or not self._key:
                        bot.exit("Buckets need AWS credentials to be created")
                    try:
                        self.bucket = self.s3.create_bucket(Bucket=self.bucket_name)
                        bot.info("Created bucket %s" % self.bucket.name)
                    except ClientError as e:
                        bot.exit(
                            "Could not create bucket {}: {}".format(
                                self.bucket_name, str(e)
                            )
                        )
                else:
                    bot.exit("Aborting due to not creating bucket")
            else:
                bot.exit("Unrecognized error code {}: {}".format(error_code, str(e)))

        return self.bucket

    def get_resource(self):
        """use the user provided endpoint and keys (from environment) to
        connect to the resource. We can share the aws environment
        variables:

        AWS_ACCESS_KEY_ID
        AWS_SECRET_ACCESS_KEY

        https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-envvars.html
        """

        # s3.ServiceResource()
        self.s3 = boto3.resource(
            "s3",
            endpoint_url=self.base,
            aws_access_key_id=self._id,
            aws_secret_access_key=self._key,
            config=boto3.session.Config(
                signature_version=UNSIGNED
                if not self._id or not self._key
                else self._signature
            ),
        )

    def _update_secrets(self, base=None):
        """update secrets will update/get the base for the server, along
        with the bucket name, defaulting to sregistry.
        """
        # We are required to have a base, either from environment or terminal
        self.base = self._get_and_update_setting("SREGISTRY_S3_BASE", self.base)
        self._id = self._get_and_update_setting("AWS_ACCESS_KEY_ID")
        self._key = self._get_and_update_setting("AWS_SECRET_ACCESS_KEY")

        if not self._id or not self._key:
            bot.warning(
                "Accessing the bucket anonymously. Consider defining AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY if access fails."
            )

        # Get the desired S3 signature.  Default is the current "s3v4" signature.
        # If specified, user can request "s3" (v2 old) signature
        self._signature = self._get_and_update_setting("SREGISTRY_S3_SIGNATURE")

        if self._signature == "s3":
            # Requested signature is S3 V2
            self._signature = "s3"
        else:
            # self._signature is not set or not set to s3 (v2), default to s3v4
            self._signature = "s3v4"

        # Define self.bucket_name, self.s3, then self.bucket
        self.get_bucket_name()
        self.get_resource()
        self.get_bucket()


Client.pull = pull
Client.push = push
Client.search = search
Client.delete = delete
Client._search_all = search_all
Client._container_search = container_search



"""

build.py: push a Singularity recipe to Singularity Registry Server to build

Copyright (C) 2019-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot, ProgressBar
from sregistry.utils import parse_image_name, remove_uri

from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

import requests
import os


def build(self, recipe, name, extra=None):
    """push a recipe file to Singularity Registry server for a Google
    Cloud (or similar) build
    """
    path = os.path.abspath(recipe)
    bot.debug("BUILD %s" % path)

    if extra is None:
        extra = {}

    if not os.path.exists(path):
        bot.exit("%s does not exist." % path)

    if not os.path.isfile(path):
        bot.exit("Build takes a Singularity recipe file")

    valid = ["google_build"]

    # Only one valid type
    if "google_build" not in extra:
        bot.exit(
            "Please include --builder google_build as the last extra arugment for Google Cloud Build"
        )

    builder_type = None
    for builder_type in extra:
        if builder_type in valid:
            break

    # Must have valid builder type
    if builder_type is None:
        bot.exit("Invalid builder type.")

    # Extract the metadata
    names = parse_image_name(remove_uri(name))

    # COLLECTION ###################################################################

    # If the registry is provided in the uri, use it
    if names["registry"] is None:
        names["registry"] = self.base

    # If the base doesn't start with http or https, add it
    names = self._add_https(names)

    # Recipe filename to upload to
    upload_to = "Singularity.%s-%s-%s" % (
        names["collection"],
        names["image"],
        names["tag"],
    )

    # Prepare build request
    url = "%s/%s/build/" % (names["registry"].replace("/api", ""), builder_type)
    SREGISTRY_EVENT = self.authorize(request_type="build", names=names)
    headers = {"Authorization": SREGISTRY_EVENT}

    bot.debug("Setting build URL to {0}".format(url))

    # Fields for build endpoint
    fields = {
        "SREGISTRY_EVENT": SREGISTRY_EVENT,
        "name": names["image"],
        "collection": names["collection"],
        "tag": names["tag"],
        "datafile": (upload_to, open(path, "rb"), "text/plain"),
    }

    encoder = MultipartEncoder(fields=fields)
    progress_callback = create_callback(encoder, self.quiet)
    monitor = MultipartEncoderMonitor(encoder, progress_callback)
    headers = {"Content-Type": monitor.content_type, "Authorization": SREGISTRY_EVENT}

    try:
        r = requests.post(url, data=monitor, headers=headers)
        r.raise_for_status()
        print("\n[Return status {0} Created]".format(r.status_code))
    except requests.HTTPError as e:
        print("\nRecipe upload failed: {0}.".format(e))
    except KeyboardInterrupt:
        print("\nRecipe upload cancelled.")


def create_callback(encoder, quiet=False):
    encoder_len = encoder.len / (1024 * 1024.0)

    bar = ProgressBar(expected_size=encoder_len, filled_char="=", hide=quiet)

    def callback(monitor):
        bar.show(monitor.bytes_read / (1024 * 1024.0))

    return callback



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import copyfile, get_file_hash, parse_image_name, remove_uri
from sqlalchemy import or_
import os
import json
import shutil


# COLLECTIONS ##################################################################


def get_or_create_collection(self, name):
    """get a collection if it exists. If it doesn't exist, create it first.

    Parameters
    ==========
    name: the collection name parsed from parse_image_names()['collection']
    """
    from sregistry.database.models import Collection

    collection = self.get_collection(name)

    # If it doesn't exist, create it
    if collection is None:
        collection = Collection(name=name)
        self.session.add(collection)
        self.session.commit()

    return collection


def get_collection(self, name):
    """get a collection, if it exists, otherwise return None."""
    from sregistry.database.models import Collection

    return Collection.query.filter(Collection.name == name).first()


def get_container(self, name, collection_id, tag="latest", version=None):
    """get a container, otherwise return None."""
    from sregistry.database.models import Container

    if not version:
        container = Container.query.filter_by(
            collection_id=collection_id, name=name, tag=tag
        ).first()
    else:
        container = Container.query.filter_by(
            collection_id=collection_id, name=name, tag=tag, version=version
        ).first()
    return container


# ACTIONS ######################################################################


def get(self, name, quiet=False):
    """Do a get for a container, and then a collection, and then return None
       if no result is found.
    Parameters
    ==========
    name: should coincide with either the collection name, or the container
          name with the collection. A query is done first for the collection,
          and then the container, and the path to the image file returned.
    """
    names = parse_image_name(remove_uri(name))

    # First look for a collection (required)
    collection = self.get_collection(name=names["collection"])
    container = None

    if collection:
        container = self.get_container(
            collection_id=collection.id,
            name=names["image"],
            tag=names["tag"],
            version=names["version"],
        )

        if container and not quiet:
            # The container image file exists [local]
            if container.image:
                print(container.image)

            # The container has a url (but not local file)
            elif container.url:
                print(container.url)
            else:
                bot.info("No storage file found for %s" % name)

    return container


def images(self, query=None):
    """List local images in the database, optionally with a query.

    Paramters
    =========
    query: a string to search for in the container or collection name|tag|uri
    """
    from sregistry.database.models import Container

    rows = []
    if query is not None:
        like = "%" + query + "%"
        containers = Container.query.filter(
            or_(
                Container.name == query,
                Container.tag.like(like),
                Container.uri.like(like),
                Container.name.like(like),
            )
        ).all()
    else:
        containers = Container.query.all()

    if len(containers) > 0:
        message = "  [date]   [client]\t[uri]"
        bot.custom(prefix="Containers:", message=message, color="RED")
        for c in containers:
            uri = c.get_uri()
            created_at = c.created_at.strftime("%B %d, %Y")
            rows.append([created_at, "   [%s]" % c.client, uri])
        bot.table(rows)
    return containers


def inspect(self, name):
    """Inspect a local image in the database, which typically includes the
    basic fields in the model.

    """
    container = self.get(name)
    if container is not None:
        collection = container.collection.name
        fields = container.__dict__.copy()
        fields["collection"] = collection
        fields["metrics"] = json.loads(fields["metrics"])
        del fields["_sa_instance_state"]
        fields["created_at"] = str(fields["created_at"])
        print(json.dumps(fields, indent=4, sort_keys=True))
        return fields
    bot.exit("image {} was not found in the database".format(name))


def rename(self, image_name, path):
    """rename performs a move, but ensures the path is maintained in storage

    Parameters
    ==========
    image_name: the image name (uri) to rename to.
    path: the name to rename (basename is taken)
    """
    container = self.get(image_name, quiet=True)

    if container:
        if container.image:
            # Derive a new filename and url in storage
            names = parse_image_name(remove_uri(path), version=container.version)
            storage = self._get_storage_name(names)
            dirname = os.path.dirname(storage)

            # This is the collection folder
            if not os.path.exists(dirname):
                os.makedirs(dirname)

            container = self.cp(move_to=storage, container=container, command="rename")

            # On successful rename of file, update the uri
            if container is not None:
                # Create the collection if doesn't exist
                collection = self.get_or_create_collection(names["collection"])
                self.session.commit()

                # Then update the container
                container = update_container_metadata(container, collection, names)
                self.session.commit()
                return container

    bot.warning("%s not found" % image_name)


def update_container_metadata(container, collection, names):
    """update container metadata, done for a move or rename. The session
    is not saved, and needs to be done so by the calling function.

    Parameters
    ==========
    container: the container object to update with names
    collection: the collection to move it to (might be the same)
    names: the result of parse_image_names
    """
    # Update all metadata for the container
    container.uri = names["uri"]
    container.name = names["image"]
    container.tag = names["tag"]
    container.collection_id = collection.id
    return container


def mv(self, image_name, path):
    """Move an image from it's current location to a new path.
    Removing the image from organized storage is not the recommended approach
    however is still a function wanted by some.

    Parameters
    ==========
    image_name: the parsed image name.
    path: the location to move the image to
    """
    container = self.get(image_name, quiet=True)

    if container is not None:
        image = container.image or ""

        # Only continue if image file exists
        if os.path.exists(image):
            # Default assume directory, use image name and path fully
            filename = os.path.basename(image)
            filedir = os.path.abspath(path)

            # If it's a file, use filename provided
            if not os.path.isdir(path):
                filename = os.path.basename(path)
                filedir = os.path.dirname(path)

            # If directory is empty, assume $PWD
            if filedir == "":
                filedir = os.getcwd()

            # Copy to the fullpath from the storage
            fullpath = os.path.abspath(os.path.join(filedir, filename))
            return self.cp(move_to=fullpath, container=container, command="move")

    bot.warning("%s not found" % image_name)


def cp(self, move_to, image_name=None, container=None, command="copy"):
    """_cp is the shared function between mv (move) and rename, and performs
    the move, and returns the updated container

    Parameters
    ==========
    image_name: an image_uri to look up a container in the database
    container: the container object to move (must have a container.image
    move_to: the full path to move it to
    """
    if not container and not image_name:
        bot.exit("A container or image_name must be provided to %s" % command)

    # If a container isn't provided, look for it from image_uri
    if not container:
        container = self.get(image_name, quiet=True)

    image = container.image or ""

    if os.path.exists(image):
        filedir = os.path.dirname(move_to)

        # If the two are the same, doesn't make sense
        if move_to == image:
            bot.exit("%s is already the name." % image)

        # Ensure directory exists
        if not os.path.exists(filedir):
            bot.exit("%s does not exist. Ensure exists first." % filedir)

        # Ensure writable for user
        if not os.access(filedir, os.W_OK):
            bot.exit("%s is not writable" % filedir)

        original = os.path.basename(image)

        try:
            shutil.move(image, move_to)
            container.image = move_to
            self.session.commit()
            bot.info("[%s] %s => %s" % (command, original, move_to))
            return container
        except:
            bot.exit("Cannot %s %s to %s" % (command, original, move_to))

    bot.warning(
        """Not found! Please pull %s and then %s to the appropriate
                   location."""
        % (container.uri, command)
    )


def rm(self, image_name):
    """Delete an image record and file from the database."""
    container = self.get(image_name)
    if container is not None:
        name = container.uri or container.get_uri()
        image = container.image
        self.session.delete(container)
        self.session.commit()
        if image is not None:
            if os.path.exists(image):
                os.remove(image)
            else:
                bot.warning(
                    "image file {} does not exist on the file system!".format(image)
                )
                return None
            return image
        bot.info("[rm] %s" % name)


def add(
    self,
    image_path=None,
    image_uri=None,
    image_name=None,
    url=None,
    metadata=None,
    save=True,
    copy=False,
):
    """get or create a container, including the collection to add it to.
    This function can be used from a file on the local system, or via a URL
    that has been downloaded. Either way, if one of url, version, or image_file
    is not provided, the model is created without it. If a version is not
    provided but a file path is, then the file hash is used.

    Parameters
    ==========
    image_path: full path to image file
    image_name: if defined, the user wants a custom name (and not based on uri)
    metadata: any extra metadata to keep for the image (dict)
    save: if True, move the image to the cache if it's not there
    copy: If True, copy the image instead of moving it.

    image_name: a uri that gets parsed into a names object that looks like:

    {'collection': 'vsoch',
     'image': 'hello-world',
     'storage': 'vsoch/hello-world-latest.img',
     'tag': 'latest',
     'version': '12345'
     'uri': 'vsoch/hello-world:latest@12345'}

    After running add, the user will take some image in a working
    directory, add it to the database, and have it available for search
    and use under SREGISTRY_STORAGE/<collection>/<container>

    If the container was retrieved from a webby place, it should have version
    If no version is found, the file hash is used.
    """

    from sregistry.database.models import Container

    # We can only save if the image is provided
    if image_path is not None:
        if not os.path.exists(image_path) and save is True:
            bot.exit("Cannot find %s" % image_path)

    # An image uri is required for version, tag, etc.
    if image_uri is None:
        bot.exit("You must provide an image uri --name <collection>/<namespace>")

    names = parse_image_name(remove_uri(image_uri))
    bot.debug("Adding %s to registry" % names["uri"])

    # If Singularity is installed, inspect image for metadata
    metadata = self.get_metadata(image_path, names=names)
    collection = self.get_or_create_collection(names["collection"])

    # Get a hash of the file for the version, or use provided
    version = names.get("version")
    if not version:
        if image_path:
            version = get_file_hash(image_path, "sha256")
        else:
            version = ""  # we can't determine a version, not in API/no file
        names = parse_image_name(remove_uri(image_uri), version=version)

    # If save, move to registry storage first
    if save and image_path:
        # If the user hasn't defined a custom name
        if image_name is None:
            image_name = self._get_storage_name(names)
        if copy:
            copyfile(image_path, image_name)
        else:
            shutil.move(image_path, image_name)
        image_path = image_name

    # Just in case the client didn't provide it, see if we have in metadata
    if url is None and "url" in metadata:
        url = metadata["url"]

    # First check that we don't have one already!
    container = self.get_container(
        name=names["image"],
        collection_id=collection.id,
        tag=names["tag"],
        version=version,
    )

    # The container did not exist, create it
    if not container:
        action = "new"
        container = Container(
            metrics=json.dumps(metadata),
            name=names["image"],
            image=image_path,
            client=self.client_name,
            tag=names["tag"],
            version=version,
            url=url,
            uri=names["uri"],
            collection_id=collection.id,
        )

        self.session.add(container)
        collection.containers.append(container)

    # The container existed, update it.
    else:
        action = "update"
        metrics = json.loads(container.metrics)
        metrics.update(metadata)
        container.url = url
        container.client = self.client_name
        if image_path is not None:
            container.image = image_path
        container.metrics = json.dumps(metrics)

    self.session.commit()
    bot.info("[container][%s] %s" % (action, names["uri"]))
    return container



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import get_file_hash, get_thumbnail, parse_image_name, remove_uri

from googleapiclient.http import MediaFileUpload
from googleapiclient.errors import HttpError
import base64
import os


def push(self, path, name, tag=None):
    """push an image to Google Cloud Drive, meaning uploading it

    path: should correspond to an absolte image path (or derive it)
    name: should be the complete uri that the user has requested to push.
    tag: should correspond with an image tag. This is provided to mirror Docker
    """
    # The root of the drive for containers (the parent folder)
    parent = self._get_or_create_folder(self._base)

    image = None
    path = os.path.abspath(path)
    bot.debug("PUSH %s" % path)

    if not os.path.exists(path):
        bot.exit("%s does not exist." % path)

    names = parse_image_name(remove_uri(name), tag=tag)
    if names["version"] is None:
        version = get_file_hash(path, "sha256")
        names = parse_image_name(remove_uri(name), tag=tag, version=version)

    # Update metadata with names, flatten to only include labels
    metadata = self.get_metadata(path, names=names)

    file_metadata = {
        "name": names["storage"],
        "mimeType": "application/octet-stream",
        "parents": [parent["id"]],
        "properties": metadata,
    }

    media = MediaFileUpload(path, resumable=True)
    try:
        bot.spinner.start()
        image = (
            self._service.files()
            .create(body=file_metadata, media_body=media, fields="id")
            .execute()
        )

        # Add a thumbnail!
        thumbnail = get_thumbnail()

        with open(thumbnail, "rb") as f:
            body = {
                "contentHints": {
                    "thumbnail": {
                        "image": base64.urlsafe_b64encode(f.read()).decode("utf8"),
                        "mimeType": "image/png",
                    }
                }
            }
            image = (
                self._service.files().update(fileId=image["id"], body=body).execute()
            )

        bot.spinner.stop()
        print(image["name"])

    except HttpError:
        bot.error("Error uploading %s" % path)

    return image



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
import datetime
import hashlib
import io
import os
import tempfile
import tarfile


def get_template(name):
    """return a default template for some function in sregistry
    If there is no template, None is returned.

    Parameters
    ==========
    name: the name of the template to retrieve

    """
    name = name.lower()
    templates = dict()

    templates["tarinfo"] = {
        "gid": 0,
        "uid": 0,
        "uname": "root",
        "gname": "root",
        "mode": 493,
    }

    if name in templates:
        bot.debug("Found template for %s" % (name))
        return templates[name]
    else:
        bot.warning("Cannot find template %s" % (name))


def create_tar(files, output_folder=None):
    """create_memory_tar will take a list of files (each a dictionary
    with name, permission, and content) and write the tarfile
    (a sha256 sum name is used) to the output_folder.
    If there is no output folde specified, the
    tar is written to a temporary folder.
    """
    if output_folder is None:
        output_folder = tempfile.mkdtemp()

    finished_tar = None
    additions = []
    contents = []

    for entity in files:
        info = tarfile.TarInfo(name=entity["name"])
        info.mode = entity["mode"]
        info.mtime = int(datetime.datetime.now().strftime("%s"))
        info.uid = entity["uid"]
        info.gid = entity["gid"]
        info.uname = entity["uname"]
        info.gname = entity["gname"]

        # Get size from stringIO write
        filey = io.StringIO()
        content = None
        try:  # python3
            info.size = filey.write(entity["content"])
            content = io.BytesIO(entity["content"].encode("utf8"))
        except:  # python2
            info.size = int(filey.write(entity["content"].decode("utf-8")))
            content = io.BytesIO(entity["content"].encode("utf8"))

        if content is not None:
            addition = {"content": content, "info": info}
            additions.append(addition)
            contents.append(content)

    # Now generate the sha256 name based on content
    if len(additions) > 0:
        hashy = get_content_hash(contents)
        finished_tar = "%s/sha256:%s.tar.gz" % (output_folder, hashy)

        # Warn the user if it already exists
        if os.path.exists(finished_tar):
            msg = "metadata file %s already exists " % finished_tar
            msg += "will over-write."
            bot.debug(msg)

        # Add all content objects to file
        tar = tarfile.open(finished_tar, "w:gz")
        for a in additions:
            tar.addfile(a["info"], a["content"])
        tar.close()

    else:
        msg = "No contents, environment or labels"
        msg += " for tarfile, will not generate."
        bot.debug(msg)

    return finished_tar


def get_content_hash(contents):
    """get_content_hash will return a hash for a list of content (bytes/other)"""
    hasher = hashlib.sha256()
    for content in contents:
        if isinstance(content, io.BytesIO):
            content = content.getvalue()
        if not isinstance(content, bytes):
            content = bytes(content)
        hasher.update(content)
    return hasher.hexdigest()



from setuptools import setup, find_packages
import codecs
import os

################################################################################
# HELPER FUNCTIONS #############################################################
################################################################################


def get_lookup():
    """get version by way of sregistry.version, returns a
    lookup dictionary with several global variables without
    needing to import singularity
    """
    lookup = dict()
    version_file = os.path.join("sregistry", "version.py")
    with open(version_file) as filey:
        exec(filey.read(), lookup)
    return lookup


# Read in requirements
def get_reqs(lookup=None, key="INSTALL_REQUIRES"):
    """get requirements, mean reading in requirements and versions from
    the lookup obtained with get_lookup"""

    if lookup == None:
        lookup = get_lookup()

    install_requires = []
    for module in lookup[key]:
        module_name = module[0]
        module_meta = module[1]
        if "exact_version" in module_meta:
            dependency = "%s==%s" % (module_name, module_meta["exact_version"])
        elif "min_version" in module_meta:
            if module_meta["min_version"] == None:
                dependency = module_name
            else:
                dependency = "%s>=%s" % (module_name, module_meta["min_version"])
        elif "max_version" in module_meta:
            dependency = "%s<=%s" % (module_name, module_meta["max_version"])
        install_requires.append(dependency)
    return install_requires


# Make sure everything is relative to setup.py
install_path = os.path.dirname(os.path.abspath(__file__))
os.chdir(install_path)

# Get version information from the lookup
lookup = get_lookup()
VERSION = lookup["__version__"]
NAME = lookup["NAME"]
AUTHOR = lookup["AUTHOR"]
AUTHOR_EMAIL = lookup["AUTHOR_EMAIL"]
PACKAGE_URL = lookup["PACKAGE_URL"]
KEYWORDS = lookup["KEYWORDS"]
DESCRIPTION = lookup["DESCRIPTION"]
LICENSE = lookup["LICENSE"]
with open("README.md") as filey:
    LONG_DESCRIPTION = filey.read()

################################################################################
# MAIN #########################################################################
################################################################################

if __name__ == "__main__":
    INSTALL_REQUIRES = get_reqs(lookup)

    # These requirement DON'T include sqlalchemy, only client

    INSTALL_BASIC_ALL = get_reqs(lookup, "INSTALL_BASIC_ALL")
    AWS_BASIC = get_reqs(lookup, "INSTALL_BASIC_AWS")
    S3_BASIC = get_reqs(lookup, "INSTALL_BASIC_S3")
    DROPBOX_BASIC = get_reqs(lookup, "INSTALL_BASIC_DROPBOX")
    REGISTRY_BASIC = get_reqs(lookup, "INSTALL_BASIC_REGISTRY")
    GLOBUS_BASIC = get_reqs(lookup, "INSTALL_BASIC_GLOBUS")
    GOOGLE_STORAGE_BASIC = get_reqs(lookup, "INSTALL_BASIC_GOOGLE_STORAGE")
    GOOGLE_DRIVE_BASIC = get_reqs(lookup, "INSTALL_BASIC_GOOGLE_DRIVE")
    GOOGLE_BUILD_BASIC = get_reqs(lookup, "INSTALL_BASIC_GOOGLE_BUILD")
    GOOGLE_COMPUTE_BASIC = get_reqs(lookup, "INSTALL_BASIC_GOOGLE_COMPUTE")
    SWIFT_BASIC = get_reqs(lookup, "INSTALL_BASIC_SWIFT")
    TESTS_REQUIRES = get_reqs(lookup, "TESTS_REQUIRES")

    # These requirement sets include sqlalchemy, for client+storage

    INSTALL_REQUIRES_ALL = get_reqs(lookup, "INSTALL_REQUIRES_ALL")
    AWS = get_reqs(lookup, "INSTALL_REQUIRES_AWS")
    S3 = get_reqs(lookup, "INSTALL_REQUIRES_S3")
    DROPBOX = get_reqs(lookup, "INSTALL_REQUIRES_DROPBOX")
    REGISTRY = get_reqs(lookup, "INSTALL_REQUIRES_REGISTRY")
    GLOBUS = get_reqs(lookup, "INSTALL_REQUIRES_GLOBUS")
    GOOGLE_STORAGE = get_reqs(lookup, "INSTALL_REQUIRES_GOOGLE_STORAGE")
    GOOGLE_DRIVE = get_reqs(lookup, "INSTALL_REQUIRES_GOOGLE_DRIVE")
    GOOGLE_COMPUTE = get_reqs(lookup, "INSTALL_REQUIRES_GOOGLE_COMPUTE")
    GOOGLE_BUILD = get_reqs(lookup, "INSTALL_REQUIRES_GOOGLE_BUILD")
    SWIFT = get_reqs(lookup, "INSTALL_REQUIRES_SWIFT")

    setup(
        name=NAME,
        version=VERSION,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        maintainer=AUTHOR,
        maintainer_email=AUTHOR_EMAIL,
        packages=find_packages(),
        include_package_data=True,
        zip_safe=False,
        url=PACKAGE_URL,
        license=LICENSE,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        keywords=KEYWORDS,
        setup_requires=["pytest-runner"],
        install_requires=INSTALL_REQUIRES,
        tests_require=TESTS_REQUIRES,
        extras_require={
            "all-basic": [INSTALL_BASIC_ALL],
            "aws-basic": [AWS_BASIC],
            "dropbox-basic": [DROPBOX_BASIC],
            "google-compute-basic": [GOOGLE_COMPUTE_BASIC],
            "google-storage-basic": [GOOGLE_STORAGE_BASIC],
            "google-drive-basic": [GOOGLE_DRIVE_BASIC],
            "google-build-basic": [GOOGLE_BUILD_BASIC],
            "globus-basic": [GLOBUS_BASIC],
            "registry-basic": [REGISTRY_BASIC],
            "s3-basic": [S3_BASIC],
            "swift-basic": [SWIFT_BASIC],
            "all": [INSTALL_REQUIRES_ALL],
            "aws": [AWS],
            "dropbox": [DROPBOX],
            "google-compute": [GOOGLE_COMPUTE],
            "google-storage": [GOOGLE_STORAGE],
            "google-drive": [GOOGLE_DRIVE],
            "google-build": [GOOGLE_BUILD],
            "globus": [GLOBUS],
            "registry": [REGISTRY],
            "s3": [S3],
            "swift": [SWIFT],
        },
        scripts=["sregistry/main/docker/blob2oci"],
        classifiers=[
            "Intended Audience :: Science/Research",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
            "Programming Language :: C",
            "Programming Language :: Python",
            "Topic :: Software Development",
            "Topic :: Scientific/Engineering",
            "Operating System :: Unix",
            "Programming Language :: Python :: 3.3",
        ],
        entry_points={"console_scripts": ["sregistry=sregistry.client:main"]},
    )



"""

Copyright (C) 2016-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""


def main(args, parser, extra):
    lookup = {"ipython": ipython, "python": python, "bpython": bpython}

    shells = ["ipython", "python", "bpython"]

    # If the user supplies a client choice, make into a uri
    if args.endpoint is not None:
        if not args.endpoint.endswith("://"):
            args.endpoint = "%s://" % args.endpoint

    from sregistry.main import get_client

    cli = get_client(args.endpoint, quiet=args.quiet)

    # If the user asked for a specific shell via environment
    shell = cli._get_and_update_setting("SREGISTRY_SHELL")
    if shell is not None:
        shell = shell.lower()
        if shell in lookup:
            try:
                return lookup[shell](args)
            except ImportError:
                pass

    # Otherwise present order of liklihood to have on system
    for shell in shells:
        try:
            return lookup[shell](args)
        except ImportError:
            pass


def ipython(args):
    """give the user an ipython shell, optionally with an endpoint of choice."""

    # The client will announce itself (backend/database) unless it's get
    from sregistry.main import get_client

    client = get_client(args.endpoint)
    client.announce(args.command)
    from IPython import embed

    embed()


def bpython(args):
    import bpython
    from sregistry.main import get_client

    client = get_client(args.endpoint)
    client.announce(args.command)
    from sregistry.database.models import Container, Collection

    bpython.embed(
        locals_={"client": client, "Container": Container, "Collection": Collection}
    )


def python(args):
    import code
    from sregistry.main import get_client

    client = get_client(args.endpoint)
    client.announce(args.command)
    from sregistry.database.models import Container, Collection

    code.interact(
        local={"client": client, "Container": Container, "Collection": Collection}
    )



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import hashlib
import errno
import os
import pwd
import shutil
import tempfile

import json
from sregistry.logger import bot


################################################################################
## FOLDER OPERATIONS ###########################################################
################################################################################


def mkdir_p(path):
    """mkdir_p attempts to get the same functionality as mkdir -p
    :param path: the path to create.
    """
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            bot.exit("Error creating path %s, exiting." % path)


def get_tmpfile(requested_tmpdir=None, prefix=""):
    """get a temporary file with an optional prefix. By default will be
    created in /tmp unless SREGISTRY_TMPDIR is set. By default, the file
    is closed (and just a name returned).

    Parameters
    ==========
    requested_tmpdir: an optional requested temporary directory, first
    priority as is coming from calling function.
    prefix: Given a need for a sandbox (or similar), prefix the file
    with this string.
    """

    # First priority for the base goes to the user requested.
    tmpdir = get_tmpdir(requested_tmpdir)

    # If tmpdir is set, add to prefix
    if tmpdir is not None:
        prefix = os.path.join(tmpdir, os.path.basename(prefix))

    fd, tmp_file = tempfile.mkstemp(prefix=prefix)
    os.close(fd)

    return tmp_file


def get_tmpdir(requested_tmpdir=None, prefix="", create=True):
    """get a temporary directory for an operation. If SREGISTRY_TMPDIR
    is set, return that. Otherwise, return the output of tempfile.mkdtemp

    Parameters
    ==========
    requested_tmpdir: an optional requested temporary directory, first
    priority as is coming from calling function.
    prefix: Given a need for a sandbox (or similar), we will need to
    create a subfolder *within* the SREGISTRY_TMPDIR.
    create: boolean to determine if we should create folder (True)
    """
    from sregistry.defaults import SREGISTRY_TMPDIR

    # First priority for the base goes to the user requested.
    tmpdir = requested_tmpdir or SREGISTRY_TMPDIR

    prefix = prefix or "sregistry-tmp"
    prefix = "%s.%s" % (prefix, next(tempfile._get_candidate_names()))
    tmpdir = os.path.join(tmpdir, prefix)

    if not os.path.exists(tmpdir) and create is True:
        os.mkdir(tmpdir)

    return tmpdir


def get_userhome():
    """get the user home based on the effective uid"""
    return pwd.getpwuid(os.getuid())[5]


################################################################################
## COMPRESSION #################################################################
################################################################################


def extract_tar(archive, output_folder, handle_whiteout=False):
    """extract a tar archive to a specified output folder

    Parameters
    ==========
    archive: the archive file to extract
    output_folder: the output folder to extract to
    handle_whiteout: use docker2oci variation to handle whiteout files

    """
    from .terminal import run_command

    # Do we want to remove whiteout files?
    if handle_whiteout is True:
        return _extract_tar(archive, output_folder)

    # If extension is .tar.gz, use -xzf
    args = "-xf"
    if archive.endswith(".tar.gz"):
        args = "-xzf"

    # Just use command line, more succinct.
    command = ["tar", args, archive, "-C", output_folder, "--exclude=dev/*"]
    if not bot.is_quiet():
        print("Extracting %s" % archive)

    return run_command(command)


def _extract_tar(archive, output_folder):
    """use blob2oci to handle whiteout files for extraction. Credit for this
    script goes to docker2oci by Olivier Freyermouth, and see script
    folder for license.

    Parameters
    ==========
    archive: the archive to extract
    output_folder the output folder (sandbox) to extract to

    """
    from .terminal import run_command, which

    result = which("blob2oci")
    if result["return_code"] != 0:
        bot.exit("Cannot find blob2oci script on path, exiting.")

    script = result["message"]
    command = ["exec", script, "--layer", archive, "--extract", output_folder]

    if not bot.is_quiet():
        print("Extracting %s" % archive)

    return run_command(command)


def get_file_hash(image_path, algorithm="sha256"):
    """return an md5 hash of the file based on a criteria level. This
    is intended to give the file a reasonable version.

    Parameters
    ==========
    image_path: full path to the singularity image

    """
    try:
        hasher = getattr(hashlib, algorithm)()
    except AttributeError:
        bot.error("%s is an invalid algorithm.")
        bot.exit(" ".join(hashlib.algorithms_guaranteed))

    with open(image_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hasher.update(chunk)
    return hasher.hexdigest()


################################################################################
## FILE OPERATIONS #############################################################
################################################################################


def copyfile(source, destination, force=True):
    """copy a file from a source to its destination."""
    # Case 1: It's already there, we aren't replacing it :)
    if source == destination and force is False:
        return destination

    # Case 2: It's already there, we ARE replacing it :)
    if os.path.exists(destination) and force is True:
        os.remove(destination)

    shutil.copyfile(source, destination)
    return destination


def write_file(filename, content, mode="w"):
    """write_file will open a file, "filename" and write content, "content"
    and properly close the file
    """
    with open(filename, mode) as filey:
        filey.writelines(content)
    return filename


def write_json(json_obj, filename, mode="w", print_pretty=True):
    """write_json will (optionally,pretty print) a json object to file

    Parameters
    ==========
    json_obj: the dict to print to json
    filename: the output file to write to
    pretty_print: if True, will use nicer formatting
    """
    with open(filename, mode) as filey:
        if print_pretty:
            filey.writelines(print_json(json_obj))
        else:
            filey.writelines(json.dumps(json_obj))
    return filename


def print_json(json_obj):
    """just dump the json in a "pretty print" format"""
    return json.dumps(json_obj, indent=4, separators=(",", ": "))


def read_file(filename, mode="r", readlines=True):
    """write_file will open a file, "filename" and write content, "content"
    and properly close the file
    """
    with open(filename, mode) as filey:
        if readlines is True:
            content = filey.readlines()
        else:
            content = filey.read()
    return content


def read_json(filename, mode="r"):
    """read_json reads in a json file and returns
    the data structure as dict.
    """
    with open(filename, mode) as filey:
        data = json.load(filey)
    return data



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import read_file, get_installdir
import os


def prepare_metadata(metadata):
    """prepare a key/value list of metadata for the request. The metadata
    object that comes in is only parsed one level.
    """
    pairs = {"metadata": {"items": [{"key": "client", "value": "sregistry"}]}}
    for key, val in metadata.items():
        if not isinstance(val, dict) and not isinstance(val, list):
            pairs["metadata"]["items"].append({"key": key, "value": val})
        elif isinstance(val, dict):
            for k, v in val.items():
                if not isinstance(v, dict) and not isinstance(v, list):
                    pairs["metadata"]["items"].append({"key": k, "value": v})

    return pairs


def get_build_template(name=None, manager="apt"):
    """get a particular build template, by default we return templates
    that are based on package managers.

    Parameters
    ==========
    name: the full path of the template file to use.
    manager: the package manager to use in the template (yum or apt)

    """
    base = get_installdir()
    if name is None:
        name = "%s/main/templates/build/singularity-builder-%s.sh" % (base, manager)

    if os.path.exists(name):
        bot.debug("Found template %s" % name)
        return "".join(read_file(name))

    bot.warning("Template %s not found." % name)



#!/usr/bin/python

# Copyright (C) 2017-2021 Vanessa Sochat.

# This Source Code Form is subject to the terms of the
# Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import shutil
import pytest


def test_write_read_files(tmp_path):
    """test_write_read_files will test the functions write_file and read_file"""
    print("Testing utils.write_file...")
    from sregistry.utils import write_file

    tmpfile = str(tmp_path / "written_file.txt")
    assert not os.path.exists(tmpfile)
    write_file(tmpfile, "hello!")
    assert os.path.exists(tmpfile)

    print("Testing utils.read_file...")
    from sregistry.utils import read_file

    content = read_file(tmpfile)[0]
    assert content == "hello!"


def test_write_bad_json(tmp_path):
    from sregistry.utils import write_json

    bad_json = {"Wakkawakkawakka'}": [{True}, "2", 3]}
    tmpfile = str(tmp_path / "json_file.txt")
    assert not os.path.exists(tmpfile)
    with pytest.raises(TypeError):
        write_json(bad_json, tmpfile)


def test_write_json(tmp_path):
    import json
    from sregistry.utils import write_json, read_json

    good_json = {"Wakkawakkawakka": [True, "2", 3]}
    tmpfile = str(tmp_path / "good_json_file.txt")
    assert not os.path.exists(tmpfile)
    write_json(good_json, tmpfile)
    with open(tmpfile, "r") as f:
        content = json.loads(f.read())
    assert isinstance(content, dict)
    assert "Wakkawakkawakka" in content
    content = read_json(tmpfile)
    assert "Wakkawakkawakka" in content


def test_check_install():
    """check install is used to check if a particular software is installed.
    If no command is provided, singularity is assumed to be the test case"""
    print("Testing utils.check_install")
    from sregistry.utils import check_install

    is_installed = check_install("echo")
    assert is_installed
    is_not_installed = check_install("fakesoftwarename")
    assert not is_not_installed


def test_get_installdir():
    """get install directory should return the base of where sregistry
    is installed
    """
    print("Testing utils.get_installdir")
    from sregistry.utils import get_installdir

    whereami = get_installdir()
    print(whereami)
    assert whereami.endswith("sregistry")


def test_remove_uri():
    print("Testing utils.remove_uri")
    from sregistry.utils import remove_uri

    assert remove_uri("docker://ubuntu") == "ubuntu"
    assert (
        remove_uri("shub://vanessa/singularity-images") == "vanessa/singularity-images"
    )
    assert remove_uri("vanessa/singularity-images") == "vanessa/singularity-images"


def ubuntu_check(names, original):
    assert names["image"] == "ubuntu"
    assert names["collection"] == "library"
    assert names["registry"] is None
    assert names["version"] is None
    assert names["tag"] == "latest"
    assert names["url"] == "library/ubuntu"
    assert names["uri"] == "library/ubuntu:latest"
    assert names["original"] == original
    assert names["storage"] == "library/ubuntu:latest.sif"


def test_parse_image_name():
    print("Testing utils.parse_image_name")
    from sregistry.utils import parse_image_name

    names = parse_image_name("ubuntu")
    ubuntu_check(names, "ubuntu")
    names = parse_image_name("ubuntu:latest")
    ubuntu_check(names, "ubuntu:latest")
    names = parse_image_name("library/ubuntu:latest")
    ubuntu_check(names, "library/ubuntu:latest")

    # Version should include same with version
    names = parse_image_name("library/ubuntu:latest@version")
    assert names["version"] == "version"
    assert names["uri"] == "library/ubuntu:latest@version"
    assert names["image"] == "ubuntu"
    assert names["collection"] == "library"
    assert names["registry"] is None
    assert names["tag"] == "latest"
    assert names["url"] == "library/ubuntu"
    assert names["uri"] == "library/ubuntu:latest@version"
    assert names["original"] == "library/ubuntu:latest@version"
    assert names["storage"] == "library/ubuntu:latest@version.sif"


def test_recipe_tag():
    print("Testing utils.get_recipe_tag")
    from sregistry.utils import get_recipe_tag

    assert get_recipe_tag("Singularity") == "latest"
    assert get_recipe_tag("/path/to/Singularity") == "latest"
    assert get_recipe_tag("/path/to/Singularity.tag") == "tag"
    assert get_recipe_tag("/path/to/Dockerfile") is None


def test_run_command():
    print("Testing utils.run_command")
    from sregistry.utils import run_command

    result = run_command(["echo", "hello"])
    assert result["message"] == "hello\n"
    assert result["return_code"] == 0


def test_get_thumbnail():
    print("Testing utils.get_thumbnail")
    from sregistry.utils import get_thumbnail

    assert os.path.exists(get_thumbnail())
    assert get_thumbnail().endswith("robot.png")


def test_which():
    print("Testing utils.which")
    from sregistry.utils import which

    result = which("echo")
    assert result["message"].endswith("echo")
    assert result["return_code"] == 0
    result = which("singularityaaaa")
    assert result["message"] == ""
    assert result["return_code"] == 1


def test_get_file_hash():
    print("Testing utils.get_file_hash")
    from sregistry.utils import get_file_hash

    here = os.path.dirname(os.path.abspath(__file__))
    testdata = os.path.join(here, "testdata", "hashtest.txt")
    assert (
        get_file_hash(testdata)
        == "6bb92117bded3da774363713657a629a9f38eac2e57cd47e1dcda21d3445c67d"
    )
    assert get_file_hash(testdata, "md5") == "e5d376ca96081dd561ff303c3a631fd5"


def test_get_uri():
    print("Testing utils.get_uri")
    from sregistry.utils import get_uri

    assert get_uri("docker://ubuntu") == "docker"
    assert "hub" in get_uri("shub://vanessa/singularity-images")
    assert get_uri("notvalid://") is None
    assert get_uri("nowvalid://", validate=False) == "nowvalid"


def test_get_userhome():
    print("Testing utils.get_userhome")
    from sregistry.utils import get_userhome

    home = get_userhome()
    assert home in os.environ.get("HOME")


def test_read_tar(tmp_path):
    print("Testing utils.extract_tar")
    from sregistry.utils import extract_tar

    here = os.path.dirname(os.path.abspath(__file__))
    testdata = os.path.join(here, "testdata", "hashtest.tar.gz")
    dirname = str(tmp_path / "destination")
    os.mkdir(dirname)
    extract_tar(testdata, dirname)
    assert os.path.exists(dirname)
    assert "hashtest.txt" in [os.path.basename(x) for x in os.listdir(dirname)]


def test_copyfile(tmp_path):
    print("Testing utils.copyfile")
    from sregistry.utils import copyfile, write_file

    original = str(tmp_path / "location1.txt")
    dest = str(tmp_path / "location2.txt")
    print(original)
    print(dest)
    write_file(original, "CONTENT IN FILE")
    copyfile(original, dest)
    assert os.path.exists(original)
    assert os.path.exists(dest)


def test_get_tmpdir_tmpfile():
    print("Testing utils.get_tmpdir, get_tmpfile")
    from sregistry.utils import get_tmpdir, get_tmpfile

    tmpdir = get_tmpdir()
    assert os.path.exists(tmpdir)
    assert os.path.basename(tmpdir).startswith("sregistry")
    shutil.rmtree(tmpdir)
    tmpdir = get_tmpdir(prefix="name")
    assert os.path.basename(tmpdir).startswith("name")
    shutil.rmtree(tmpdir)
    tmpfile = get_tmpfile()
    assert "sregistry" in tmpfile
    os.remove(tmpfile)
    tmpfile = get_tmpfile(prefix="pancakes")
    assert "pancakes" in tmpfile
    os.remove(tmpfile)


def test_mkdir_p(tmp_path):
    print("Testing utils.mkdir_p")
    from sregistry.utils import mkdir_p

    dirname = str(tmp_path / "input")
    result = os.path.join(dirname, "level1", "level2", "level3")
    mkdir_p(result)
    assert os.path.exists(result)


def test_print_json():
    print("Testing utils.print_json")
    from sregistry.utils import print_json

    result = print_json({1: 1})
    assert result == '{\n    "1": 1\n}'



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.auth import get_secrets_file, read_client_secrets
from sregistry.logger import bot
from sregistry.utils import write_json
import json


def main(args, parser, extra):
    # No commands provided, print help, show clients

    if len(args.commands) == 0:
        usage()
        bot.exit("You must provide a valid command")

    # Get the chosen action
    command = args.commands.pop(0)
    backend = None

    # If still one more, there is a targeted backend
    if len(args.commands) > 0:
        backend = args.commands.pop(0)

    # Option 1: The user wants to list all or a specific command
    if command == "ls":
        list_backends(backend)

    # Option 2: Remove a configuration
    elif command == "delete":
        delete_backend(backend)

    # Option 3: Activate a backend
    elif command == "activate":
        activate(backend)

    # Option 4: Deactivate a backend
    elif command == "deactivate":
        deactivate()

    # Option 5: get a status

    elif command == "status":
        status(backend)

    # Option 6: add a config value

    elif command == "add":
        if len(args.commands) > 1:
            variable = args.commands.pop(0)
            value = args.commands.pop(0)
            add(backend, variable, value, force=args.force)
        else:
            print("Usage: sregistry backend add <backend> <variable> <value>")

    # Option 7: Remove a config value
    elif command in ["remove", "rm"]:
        if len(args.commands) > 0:
            variable = args.commands.pop(0)
            remove(backend, variable)
        else:
            print("Usage: sregistry backend remove <backend> <variable>")

    else:
        bot.exit("%s is not a recognized command." % command)


def usage():
    print(
        """
             sregistry backend ls:     list backends found in secrets
             sregistry backend ls hub
             sregistry backend status: get status
             sregistry backend rm <backend> remove a backend
             sregistry backend de|activate: activate or deactivate
             sregistry backend remove nvidia token
             sregistry backend add nvidia token 123455
               
             SREGISTRY_NVIDIA_TOKEN 123455
          """
    )


def status(backend):
    """print the status for all or one of the backends."""
    print("[backend status]")
    settings = read_client_secrets()
    print("There are %s clients found in secrets." % len(settings))
    if "SREGISTRY_CLIENT" in settings:
        print("active: %s" % settings["SREGISTRY_CLIENT"])
        update_secrets(settings)
    else:
        print("There is no active client.")


def add(backend, variable, value, force=False):
    """add the variable to the config"""
    print("[add]")
    settings = read_client_secrets()

    # If the variable begins with the SREGISTRY_<CLIENT> don't add it
    prefix = "SREGISTRY_%s_" % backend.upper()
    if not variable.startswith(prefix):
        variable = "%s%s" % (prefix, variable)

    # All must be uppercase
    variable = variable.upper()
    bot.info("%s %s" % (variable, value))

    # Does the setting already exist?

    if backend in settings:
        if variable in settings[backend] and force is False:
            previous = settings[backend][variable]
            bot.exit(
                "%s is already set as %s. Use --force to override."
                % (variable, previous)
            )

    if backend not in settings:
        settings[backend] = {}

    settings[backend][variable] = value
    update_secrets(settings)


def remove(backend, variable):
    """remove a variable from the config, if found."""
    print("[remove]")
    settings = read_client_secrets()

    # If the variable begins with the SREGISTRY_<CLIENT> don't add it
    prefixed = variable
    prefix = "SREGISTRY_%s_" % backend.upper()
    if not variable.startswith(prefix):
        prefixed = "%s%s" % (prefix, variable)

    # All must be uppercase
    variable = variable.upper()
    bot.info(variable)

    # Does the setting already exist?
    if backend in settings:
        if variable in settings[backend]:
            del settings[backend][variable]
        if prefixed in settings[backend]:
            del settings[backend][prefixed]
        update_secrets(settings)


def activate(backend):
    """activate a backend by adding it to the .sregistry configuration file."""
    settings = read_client_secrets()
    if backend is not None:
        settings["SREGISTRY_CLIENT"] = backend
        update_secrets(settings)
        print("[activate] %s" % backend)


def update_secrets(secrets):
    secrets_file = get_secrets_file()
    write_json(secrets, secrets_file)


def delete_backend(backend):
    """delete a backend, and update the secrets file"""
    settings = read_client_secrets()
    if backend in settings:
        del settings[backend]

        # If the backend was the active client, remove too
        if "SREGISTRY_CLIENT" in settings:
            if settings["SREGISTRY_CLIENT"] == backend:
                del settings["SREGISTRY_CLIENT"]

        update_secrets(settings)
        print("[delete] %s" % backend)
    else:
        if backend is not None:
            print("%s is not a known client." % backend)
        else:
            print("Please specify a backend to delete.")


def deactivate():
    """deactivate any active backend by removing it from the .sregistry
    configuration file
    """
    settings = read_client_secrets()
    if "SREGISTRY_CLIENT" in settings:
        del settings["SREGISTRY_CLIENT"]
        update_secrets(settings)
        print("[deactivated]")
    else:
        print("There is no active client")


def list_backends(backend=None):
    """return a list of backends installed for the user, which is based on
    the config file keys found present

    Parameters
    ==========
    backend: a specific backend to list. If defined, just list parameters.

    """
    settings = read_client_secrets()

    # Backend names are the keys
    backends = list(settings.keys())
    backends = [b for b in backends if b != "SREGISTRY_CLIENT"]

    if backend in backends:
        bot.info(backend)
        print(json.dumps(settings[backend], indent=4, sort_keys=True))
    else:
        if backend is not None:
            print("%s is not a known client." % backend)
        bot.info("Backends Installed")
        print("\n".join(backends))



"""

Copyright (C) 2018-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import remove_uri

import sys


def search(self, query=None, args=None):
    """query a Singularity registry for a list of images.
    If query is None, collections are listed.

    """

    if query is not None:
        # Here you might do a function that is a general list
        # Note that this means adding the function Client in __init__
        return self._container_query(query)

    # or default to listing (searching) all things.
    return self._search_all()


def search_all(self):
    """a "show all" search that doesn't require a query"""

    results = set()

    # Here we get names of collections, and then look up containers
    for container in self.conn.get_account()[1]:
        # The result here is just the name
        for result in self.conn.get_container(container["name"])[1]:
            results.add("%s/%s" % (container["name"], result["name"]))

    if len(results) == 0:
        bot.info("No container collections found.")
        sys.exit(1)

    bot.info("Collections")
    bot.table([[x] for x in list(results)])
    return list(results)


def container_query(self, query):
    """search for a specific container.
    This function would likely be similar to the above, but have different
    filter criteria from the user (based on the query)
    """

    results = set()

    query = remove_uri(query)

    # Here we get names of collections, and then look up containers
    for container in self.conn.get_account()[1]:
        # The result here is just the name
        for result in self.conn.get_container(container["name"])[1]:
            if query in result["name"]:
                results.add("%s/%s" % (container["name"], result["name"]))

    if len(results) == 0:
        bot.info("No container collections found.")
        sys.exit(1)

    bot.info("Collections")
    bot.table([list(results)])
    return list(results)



"""

Copyright (C) 2016-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot


def main(args, parser, extra):
    from sregistry.main import get_client

    image = args.query
    cli = get_client(image, quiet=args.quiet)

    # If the client doesn't have the command, exit
    if not hasattr(cli, "get"):
        bot.exit("get for images requires using the sqlite database.")

    path = cli.get(image)
    if path is None:
        bot.exit("image {} not found in database".format(image))



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
import re


# Regular expressions to parse registry, collection, repo, tag and version
_docker_uri = re.compile(
    "(?:(?P<registry>[^/@]+[.:][^/@]*)/)?"
    "(?P<collection>(?:[^:@/]+/)+)?"
    "(?P<repo>[^:@/]+)"
    "(?::(?P<tag>[^:@]+))?"
    "(?:@(?P<version>.+))?"
    "$"
)

# Reduced to match registry:port/repo or registry.com/repo
_reduced_uri = re.compile(
    "(?:(?P<registry>[^/@]+[.:][^/@]*)/)?"
    "(?P<repo>[^:@/]+)"
    "(?::(?P<tag>[^:@]+))?"
    "(?:@(?P<version>.+))?"
    "$"
    "(?P<collection>.)?"
)

# Default
_default_uri = re.compile(
    "(?:(?P<registry>[^/@]+)/)?"
    "(?P<collection>(?:[^:@/]+/)+)"
    "(?P<repo>[^:@/]+)"
    "(?::(?P<tag>[^:@]+))?"
    "(?:@(?P<version>.+))?"
    "$"
)


def set_default(item, default, use_default):
    """if an item provided is None and boolean use_default is set to True,
    return the default. Otherwise, return the item.
    """
    if item is None and use_default:
        return default
    return item


def parse_image_name(
    image_name,
    tag=None,
    version=None,
    defaults=True,
    ext="sif",
    default_collection="library",
    default_tag="latest",
    base=None,
    lowercase=True,
):
    """return a collection and repo name and tag
    for an image file.

    Parameters
    =========
    image_name: a user provided string indicating a collection,
                image, and optionally a tag.
    tag: optionally specify tag as its own argument
         over-rides parsed image tag
    defaults: use defaults "latest" for tag and "library"
              for collection.
    base: if defined, remove from image_name, appropriate if the
          user gave a registry url base that isn't part of namespace.
    lowercase: turn entire URI to lowercase (default is True)
    """

    # Save the original string
    original = image_name

    if base is not None:
        image_name = image_name.replace(base, "").strip("/")

    # If a file is provided, remove extension
    image_name = re.sub("[.](img|simg|sif)", "", image_name)

    # Parse the provided name
    uri_regexes = [_reduced_uri, _default_uri, _docker_uri]

    for r in uri_regexes:
        match = r.match(image_name)
        if match:
            break

    if not match:
        bot.exit('Could not parse image "%s"! Exiting.' % image_name)

    # Get matches
    registry = match.group("registry")
    collection = match.group("collection")
    repo_name = match.group("repo")
    repo_tag = tag or match.group("tag")
    version = version or match.group("version")

    # A repo_name is required
    assert repo_name

    # If a collection isn't provided
    collection = set_default(collection, default_collection, defaults)
    repo_tag = set_default(repo_tag, default_tag, defaults)

    # The collection, name must be all lowercase
    if lowercase:
        collection = collection.lower().rstrip("/")
        repo_name = repo_name.lower()
        repo_tag = repo_tag.lower()
    else:
        collection = collection.rstrip("/")

    if version is not None:
        version = version.lower()

    # Piece together the uri base
    if registry is None:
        uri = "%s/%s" % (collection, repo_name)
    else:
        uri = "%s/%s/%s" % (registry, collection, repo_name)

    url = uri

    # Tag is defined
    if repo_tag is not None:
        uri = "%s:%s" % (url, repo_tag)

    # Version is defined
    storage_version = None
    if version is not None:
        uri = "%s@%s" % (uri, version)
        storage_version = "%s.%s" % (uri, ext)

    # A second storage URI honors the tag (:) separator

    storage = "%s.%s" % (uri, ext)
    result = {
        "collection": collection,
        "original": original,
        "registry": registry,
        "image": repo_name,
        "url": url,
        "tag": repo_tag,
        "version": version,
        "storage": storage_version or storage,
        "uri": uri,
    }

    return result


def get_uri(image, validate=True):
    """get the uri for an image, if within acceptable

    Parameters
    ==========
    image: the image uri, in the format <uri>://<registry>/<namespace>:<tag>
    validate: if True, check if uri is in list of supported (default True)

    """
    # Ensure we have a string
    image = image or ""

    # Find uri prefix, including ://
    regexp = re.compile("^.+://")
    uri = regexp.match(image)

    if uri is not None:
        uri = uri.group().lower().replace("_", "-").replace("://", "")

        accepted_uris = [
            "aws",
            "docker",
            "http",
            "https",  # Must be allowed for pull
            "dropbox",
            "gitlab",
            "globus",
            "google-build",
            "google-storage",
            "google-drive",
            "hub",
            "nvidia",
            "registry",
            "s3",
            "swift",
        ]

        # Allow for Singularity compatability
        if "shub" in uri:
            uri = "hub"

        if validate is True and uri not in accepted_uris:
            bot.warning("%s is not a recognized uri." % uri)
            uri = None

    return uri


def remove_uri(image):
    """remove_uri will remove the uri from the image path, if provided.

    Parameters
    ==========
    image: the image uri, in the format <uri>://<registry>/<namespace>:<tag>

    """
    return re.sub("^.+://", "", image)


def get_recipe_tag(path):
    """get a recipe tag (the extension of a Singularity file). The extension
    determines the tag. If no extension is found, latest is used.

    Parameters
    ==========
    path: the path to the recipe file (e.g. /opt/Singularity.tag)

    """
    tag = None
    if re.search("Singularity", path):
        tag = re.sub("(.+)?Singularity[.]?", "", path)
        if tag in ["", None]:
            tag = "latest"
    return tag



"""

Copyright (C) 2018-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
import requests
import sys

try:
    from urllib.parse import quote_plus  # python 3.*
except:
    from urllib import quote_plus  # python 2.*


def search(self, query=None, args=None):
    """query a GitLab artifacts folder for a list of images.
    If query is None, collections are listed.
    """
    if query is None:
        bot.exit("You must include a collection query, <collection>/<repo>")

    # or default to listing (searching) all things.
    return self._search_all(query)


def search_all(self, collection, job_id=None):
    """a "show all" search that doesn't require a query
    the user is shown URLs to
    """

    results = [["job_id", "browser"]]

    url = "%s/projects/%s/jobs" % (self.api_base, quote_plus(collection.strip("/")))

    response = requests.get(url, headers=self.headers)
    if response.status_code == 200:
        jobs = response.json()

        # We can't get a listing of artifacts
        # https://gitlab.com/gitlab-org/gitlab-ce/issues/51515
        # Parse through jobs (each can have different tags for a collection):
        for job in jobs:
            # Only show jobs that are successful
            if job["status"] == "success":
                name = job["name"]

                for artifact in job["artifacts"]:
                    if artifact["filename"].endswith("zip"):
                        # The user must browse to see the names
                        artifact_url = "%s/%s/-/jobs/%s/artifacts/browse/%s" % (
                            self.base,
                            collection,
                            job["id"],
                            name,
                        )
                        results.append([str(job["id"]), artifact_url])

    if len(results) == 1:
        bot.info("No potential archives found in artifacts.")
        sys.exit(0)

    bot.info("Artifact Browsers (you will need path and job id for pull)")
    bot.table(results)
    return results
